'''Tool for creating and editing clusters with the K-Means method
'''
import http.server
import os
import socketserver
import webbrowser
from json import dumps, load
from math import floor
from shutil import copytree
from tempfile import mkdtemp

import numpy as np
from pyproj import Proj
from sklearn.cluster import KMeans


def create_clusters(locations, n_clusters):
    """Main function. Creates the clusters and calls the web browser
       to edit the result

    Args:
        locations (list, str): The locations list or the file path with the
                               json data
        n_clusters (int): The number of clusters to create
    """
    if isinstance(locations, str):
        f_p = open(locations)
        locations = load(f_p)

    positions = np.zeros([len(locations), 2])
    utm = calculate_utm_def([locations[0]['lon'], locations[0]['lat']])
    positions_list = []
    for i, loc_value in enumerate(locations):
        positions[i][0] = loc_value['lon']
        positions[i][1] = loc_value['lat']
        value = utm(loc_value['lon'], loc_value['lat'])
        positions_list.append(value)
    positions_list = np.array(positions_list)
    kmeans = KMeans(n_clusters=n_clusters, random_state=0).fit(positions_list)

    out_geojson = {"type": "FeatureCollection",
                   "features": []}
    for i, labels_value in enumerate(kmeans.labels_):
        locations[i]['cluster'] = labels_value
        out_geojson['features'].append({
            "type": "Feature",
            "geometry": {
                "type": "Point",
                "coordinates": [locations[i]['lon'], locations[i]['lat']]
            },
            "properties": {
                "cluster": int(labels_value),
                "id": locations[i]['id'],
                "alt": locations[i]['alt']
            }
        })

    web_dir = mkdtemp()
    copytree("./cluster/web", web_dir+"/web")
    f_p = open(web_dir+"/web/points.json", "w")
    f_p.write(dumps(out_geojson))
    f_p.close()

    os.chdir(web_dir + "/web")
    socketserver.socket.setdefaulttimeout(5)
    Handler = http.server.SimpleHTTPRequestHandler
    httpd = socketserver.TCPServer(("", 8000), Handler)
    print("serving at port", 8000)
    webbrowser.get('firefox').open_new_tab("localhost:8000")

    httpd.serve_forever()


def calculate_utm_def(point):
    """Returns the proj4 definition, calculating the utm zone from
    a point definition

    Args:
        point (list): A two element list with the lon and lat values
                      from a point

    Returns:
        object: The pyproj.Proj object to use to convert
    """

    zone = (floor((point[0] + 180)/6) % 60) + 1
    south = " +south " if point[1] < 0 else " "
    desc = ("+proj=utm +zone={}{}+ellps=WGS84 +datum=WGS84 " +
            "+units=m +no_defs").format(zone, south)
    return Proj(desc)



"""
Inverse distance weighting (IDW)

Compute the score of query points based on the scores of their k-nearest
neighbours, weighted by the inverse of their distances.

For more information, see :ref:`Inverse of the distance weighting - IDW`

Credits:
https://en.wikipedia.org/wiki/Inverse_distance_weighting
"""

import numpy as np
from scipy.spatial import cKDTree


class Tree(object):
    """
    Compute the score of query points based on the scores of their k-nearest
    neighbours, weighted by the inverse of their distances.
    """

    def __init__(self, coordinates=None, scores=None, leafsize=10):
        """
        Args:
            coordinates (ndarray, optional): Defaults to None.
                                             Coordinates of N sample points
                                             in a d-dimensional space.
            scores (ndarray, optional): Defaults to None.
                                        Corresponding scores.
            leafsize (int, optional): Defaults to 10.
                                      Leafsize of KD-tree data structure;
                                      should be less than 20.
        """

        if coordinates is not None:
            self.tree = cKDTree(coordinates, leafsize=leafsize)
        if scores is not None:
            self.scores = scores

    def fit(self, coordinates=None, scores=None, leafsize=10):
        """Instantiate KDtree for fast query of k-nearest neighbour distances.

        Args:
            coordinates ((N, d) ndarray, optional): Defaults to None.
                         Coordinates of N sample points in a d-dimensional
                         space.
            scores ((N,) ndarray, optional): Defaults to None. Corresponding
                                             scores.
            leafsize (int, optional): Defaults to 10.
                                      Leafsize of KD-tree data structure;
                                      should be less than 20.

        Returns:
            object: idw_tree instance

            """
        return self.__init__(coordinates, scores, leafsize)

    def __call__(self, coordinates, num_nearest=6, eps=1e-6, p_norm=2,
                 regularize_by=1e-9):
        """
        Compute the score of query points based on the scores of their
        k-nearest neighbours, weighted by the inverse of their distances.

        Args:
            coordinates ((N, d) ndarray):
                Coordinates of N query points in a d-dimensional space.
            num_nearest (int): Default to 6
                Number of nearest neighbours to use.
            p_norm (int or inf):
                Which Minkowski p-norm to use.
                1 is the sum-of-absolute-values "Manhattan" distance
                2 is the usual Euclidean distance
                infinity is the maximum-coordinate-difference distance
            eps (float): Defaults to 1e-6
                Return approximate nearest neighbors; the k-th returned value
                is guaranteed to be no further than (1+eps) times the
                distance to the real k-th nearest neighbor.
            regularize_by (float): (default 1e-9)
                Regularize distances to prevent division by zero
                for sample points with the same location as query points.

        Returns:
            (N,) ndarray: Corresponding scores.
        """
        distances, idx = self.tree.query(coordinates, num_nearest,
                                         eps=eps, p=p_norm)
        distances += regularize_by
        weights = self.scores[idx.ravel()].reshape(idx.shape)
        m_w = np.sum(weights/distances, axis=1) / np.sum(1./distances,
                                                         axis=1)
        return m_w

    def transform(self, coordinates, num_nearest=6,
                  p_norm=2, eps=1e-6, regularize_by=1e-9):
        """Compute the score of query points based on the scores of their
        k-nearest neighbours, weighted by the inverse of their distances.

        Args:
            coordinates ((N, d) ndarray):
                Coordinates of N query points in a d-dimensional space.
            num_nearest (int): Defaults to 6
                Number of nearest neighbours to use.
            p_norm (int or inf):
                Which Minkowski p-norm to use.
                1 is the sum-of-absolute-values "Manhattan" distance
                2 is the usual Euclidean distance
                infinity is the maximum-coordinate-difference distance
            eps (float): Defaults to 1e-6
                Return approximate nearest neighbors; the k-th returned value
                is guaranteed to be no further than (1+eps) times the
                distance to the real k-th nearest neighbor.
            regularize_by (float): Defaults to 1e-9
                Regularize distances to prevent division by zero
                for sample points with the same location as query points.
        Returns:
            (N,) ndarray: Corresponding scores.

        """
        return self.__call__(coordinates, num_nearest, eps,
                             p_norm, regularize_by)


def idw(residues, size, geotransform, num_nearest=6):
    '''Interpolates the residues field using the inverse of the distance
        weighting method

    Args:
        residues (dict): The residues dict
        size (list): x X y
        geotransform (list): The geotransform to apply to relate the
                                residues coordinates and the position in
                                the matrix.
                                See https://www.gdal.org/gdal_datamodel.html
                                for more information
    Returns:
        list: The interpolated residues

    '''
    coords = []
    values = []
    for key in residues.keys():
        coords.append([residues[key]['x'], residues[key]['y']])
        values.append(residues[key]['value'])
    coords = np.array(coords)
    values = np.array(values)

    idw_tree = Tree(coords, values)

    x_coords = np.arange(geotransform[0],
                         geotransform[0]+size[1]*geotransform[1],
                         geotransform[1])

    y_coords = np.arange(geotransform[3],
                         geotransform[3]+size[0]*geotransform[5],
                         geotransform[5])

    xy_coords = np.meshgrid(x_coords, y_coords)
    xy_coords = np.reshape(xy_coords, (2, -1)).T

    idw_array = idw_tree(xy_coords, num_nearest)
    idw_array = idw_array.reshape(size)

    return idw_array
